

signature POSTCOMPOSE = sig
    val postcompose : TTS.TTS -> TTS.TTS
end;
 


structure PostCompose : POSTCOMPOSE = struct


(* attaches time constraints to tts transitions *)
(* for each transition: get the port used in the transition, if any, then store its time interval within the transition *)
fun insertTime (ProcessDecl (b,c,d,e,_,k,f,h,i,init,g)) =
    let fun getPortName s1 =
	    case s1 of
		StSequence(a1,a2) => (case getPortName a1 of  "" => getPortName a2 | st => st)
	      | StSignal pn=> pn
	      | StInput(pn,_,_)=> pn
	      | StInputList(pn,_) => pn
	      | StOutput(pn,_,_)=> pn
	      | _ => ""
	fun updateBound (tag,st,stm,bnd,prl) = 
            let val port = getPortName stm
	    in case List.find (fn (x,_,_) => x=port) h of SOME (_,_,tc) => (tag,st,stm,tc,prl) | NONE => (tag,st,stm,bnd,prl)
            end
    in ProcessDecl (b,c,d,e,[],k,f,[],i,init,map updateBound g) handle BYE => raise BYE | _ => failwith "insertTime" 
    end
  | insertTime d = d;



(* replaces communication statements by null, after composition *)
(* also concats tags into a single name for the transition *)
(* also builds priority relation, replacing elem tags by transition names *)
fun removePorts (ProcessDecl (b,c,d,e,_,k,f,h,i,init,trl)) =
    let val L = ref ""
	(* unify patterns, returns nuber of solutions (0, 1, or 2 meaning "more than 1") *)
	fun unify_patterns stll =
	    let fun isConst p =
		    case p of
			BoolExp _ => true
		      | IntExp _ => true
		      | EnumExp _ => true
		      | ConExp _ => true
		      | ConstrExp (_,p,_) => isConst p
		      | _ => false
		exception NO
		fun unify p1 p2 =
		    case (p1,p2) of
			(BoolExp b1,BoolExp b2) => if b1 = b2 then p1 else raise NO
		      | (IntExp (i1,_),IntExp (i2,_)) => if i1 = i2 then p1 else raise NO
		      | (EnumExp (c1,_),EnumExp (c2,_)) => if c1 = c2 then p1 else raise NO
		      | (ConExp (c1,_),ConExp (c2,_)) => if c1 = c2 then p1 else raise NO
		      | (ConstrExp (c1,p1,ty), ConstrExp(c2,p2,_)) => if c1 = c2 then ConstrExp (c1,unify p1 p2,ty) else raise NO
		      | (VarExp _,_) => p2
		      | (_,VarExp _) => p1
		      | (AnyExp l1, AnyExp l2) => AnyExp (l1@l2)
		      | (AnyExp _,_) => p2
		      | (_,AnyExp _) => p1
		      | _ => AnyExp []
		fun list_unify [h] = h
		  | list_unify (h::t) = map (fn (p1,p2) => unify p1 p2) (ListPair.zip (h,list_unify t))
		val pl = list_unify stll
	    in (if List.all isConst pl then 1 else 2,pl)
	    end
	    handle NO => (0,[])
	fun rmIO s =
	    case s of
		StSignal p => (L := p; StNull)
	      | StOutput (p,_,expl) =>
	        if List.exists (fn e => case e of AnyExp (_::_) => true | _ => false) expl then
		  (* unsupported yet *) comperror ("unsupported open input on port "^p)
		else
	          (L := p; StNull)
	      (* | StOutput (p,_,_) => (L := p; StNull) *)
	      | StInput (p,_,stl) => rmIO (StInputList (p,[stl]))
	      | StInputList (p,stll) =>
		(* computes unification of all patterns; reject if more than one solution *)
		(* Note: works only if all conditions captured by patterns, not "where" clauses *)
		let val (k,pl1) = unify_patterns stll
		in case k of
		       0 => (L := p; StGuard (BoolExp false))  (* handled; will be removed *)
		     | 1 => (L := p;  (* handled; will be kept or removed *)
			     List.foldr StSequence StNull (flat (map (fn stList => map (fn (x,y) => ToTTS.convAssignMatch(StAssign(x,y))) (ListPair.zip (stList,pl1))) stll)))
		     | 2 => (* unsupported yet *) comperror ("unsupported open input on port "^p)
		end
	      | _ => s
	(* maps elementary tags to transition names *)
	(* should use a tree rather than an a-list ? *)
	val dict = ref []
	fun record tg name =
	    case List.find (fn (g,_) => g=tg) (!dict) of
		SOME (_,r) => r := name :: !r | _ => dict := (tg,ref [name]) :: !dict
	fun get tg =
	    let val ref l = assoc tg (!dict) in l end
        (* first pass: remove ports, build transition names and record them in dict *)
	val trl' = map (fn ((tags,lab),fr,stm,bnd,prl) =>
			   let val trname = String.concatWith "_" tags
			   in List.app (fn tg => record tg trname) tags;
			      (([trname],lab),fr,mapStm rmIO stm,bnd,prl)
			   end) trl
        (* second pass: update priority relation (replace tags by transition names) *)
	val trl'' = map (fn (namelab,fr,stm,bnd,prl) => (namelab,fr,stm,bnd,makeset (flat (map get prl)))) trl'
    in ProcessDecl (b,c,d,e,[],k,f,h,i,init,trl'')
       handle BYE => raise BYE | _ => failwith "removePorts"
    end
  | removePorts d = d;
                                                                                          



(* moves conditional statements after assignements to front. *)
(* general form (S is state):
      S := f S; if c S then st end ==> if c (f S) then S := f S; st end
   example 1: (independant)
      Z := 3; if X>Y then s end ==> if X>Y then Z := 3; s end
   example 2: (not independant), with analysis: (not implemented)
      a[z]:=3; if a[x]>a[y] then s end ==> if ((x=z) ? 3 : a[x]) > ((y=z) ? 3 : a[y]) then a[z] := 3; s end
   example 2.2: (not independant), with state copying:
      a[z]:=3; if a=b then s end ==> if (a[z]:=3;a=b) then s end
   example 3: (not independant but substitutable)
      X := Y+3; if X>Y then s end ==> if Y+3>Y then X := Y+3; s end
*)
fun normGuards (ProcessDecl (b,c,d,e,_,k,f,h,i,init,g)) =
    let fun swap s =
	    let fun simplePatt pl = List.all (fn VarExp _ => true | _ => false) pl
		fun swapGuard s =
		    case s of
			StSequence (StGuard (BoolExp true),s2) => swapGuard s2
		      | StSequence (StGuard c1,s2) =>
			(case swapGuard s2 of
			     StSequence(StGuard (BoolExp true), s2') => StSequence(StGuard c1, s2')
			   | StSequence(StGuard c2, s2') => StSequence (StGuard (InfixExp(AND,BoolType,c1,c2)), s2'))
		      | StSequence (StNull,s2) => swapGuard s2
		      | StSequence (s1 as StAssign(p,e),s2) =>
			if equal p e then swapGuard s2  else
			let val StSequence(StGuard c2,s2') = swapGuard s2
			    val ws1 = writtenInSt s1
			    val rc = readInExp c2
			    val rs1 = readInSt s1
			    val wc = writtenInExp c2
			in if List.all (fn v => not (mem v wc)) rs1 andalso
			      List.all (fn v => not (mem v ws1)) rc andalso
			      List.all (fn v => not (mem v wc)) ws1
			   then (* independant  *) StSequence (StGuard c2, StSequence(s1,s2'))
			   else (* *)
			        if (* (R(s1) union W(s1)) inter W(c2) empty *)
				   List.all (fn v => not (mem v wc)) (rs1 @ ws1) andalso
				   not (matchExp (fn LetExp _ => true | _ => false) c2) andalso
				   (* p is a variable, not an access ! *)
				   (case p of VarExp _ => true | _ => false) andalso    
				   (* no function in e [may be costly otherwise] *)
				   not (matchExp (fn AppExp _ => true | CallExp _ => true | _ => false) e)
				then (* substitutable *)
				     StSequence (StGuard (mapExp (fn x => if equal p x then e else x) c2), StSequence(s1,s2'))
				else (* *)
				     (* general case  *)
				     StSequence (StGuard (LetExp(s1,c2)), StSequence(s1,s2'))
			end
		      | StSequence (s1,s2) => (* s1 is foreach, while, if *)
			let val StSequence(StGuard c2,s2') = swapGuard s2
			    val ws1 = writtenInSt s1
			    val rc = readInExp c2
			    val rs1 = readInSt s1
			    val wc = writtenInExp c2
			in if List.all (fn v => not (mem v wc)) rs1 andalso
			      List.all (fn v => not (mem v ws1)) rc andalso
			      List.all (fn v => not (mem v wc)) ws1
			   then (* independant  *) StSequence (StGuard c2, StSequence(s1,s2'))
			   else (* general case *) StSequence (StGuard (LetExp(s1,c2)), StSequence(s1,s2'))
			end
		      | _ => StSequence (StGuard (BoolExp true),s)
	    in swapGuard (#1 (ToTTS.normSequence s))  (* normSequence needed for e.g. abp.fcr, due to removePorts introducing StNull ? *)
	    end
    in ProcessDecl (b,c,d,e,[],k,f,h,i,init,map (fn (tg,fr,s,b,p) => (tg,fr,swap s,b,p)) g)
       handle BYE => raise BYE | _ => failwith "normGuards"
    end
 | normGuards d = d;



(* analyzes usage of variables:
  - removes unused variables
  - makes temporaries assigned and part of state local variables
  - makes local variables not part of state temporary variables
  - inlines temporaries not assigned and of basic types
 A variable is part of state iff
  - some transition reads it before writing it (if it does)
  - some (possibly the same) transition writes it
 also:
  - simplifies expressions after inlining
  - removes trivially dead transitions
*)
val DEADTRANS = ref ([] : (string * string list ref) list);  (* tag and prio relation of dead trans, used in simplify and reduce *)
fun simplify (d as ProcessDecl (name,p,a,states,_,tmpvars,locvars,locports,prio,init,trl)) =
    let val REMOVED = ref false (* true if some useless assignement(s) removed *)
	fun moveVariables (ProcessDecl (name,p,a,states,rtmpvars,tmpvars,locvars,locports,prio,init,trl)) =
	    let (* encode usages of variables in process.
		   State means "read by t (before being written in t) and written in t' (possibly t=t') *)
		(* rtmpvars : readonly variables that could not be inlined (structures);
		   tmpvars  : read/write variables not part of state;
                   locvars  : variables part of state
		 *)
		datatype Usage = Unused | Read | Write | WritePartial | WriteRead | State
		(* accu accumulate usages of variables in all transitions *)
		val vars = rtmpvars @ tmpvars @ locvars
		val accu = ref (map (fn (s,_,_) => (s,Unused)) vars)
		fun merge all = (*  v has State if one transition at least reads it (before writing it)
		                    and some transition writes it (possibly the same) *)
		    accu := List.map (fn (N as (s,new),O as (_,old)) =>
					 case (new,old) of (_             ,Unused)       => (s,new)
							 | (State         ,_)            => (s,State)
							 | (Write         ,Read)         => (s,State)
							 | (WritePartial  ,Read)         => (s,State)
							 | (WriteRead     ,Read)         => (s,State)
							 | (WriteRead     ,Write)        => (s,WriteRead)
							 | (Read          ,Write)        => (s,State)
							 | (Read          ,WritePartial) => (s,State)
							 | (WriteRead     ,WritePartial) => (s,WriteRead)
							 | (Read          ,WriteRead)    => (s,State)
							 | _                             => O) (ListPair.zip (all,!accu))
		(* updates usage info for s within a transition *)
		fun uupdate v [] new = []
		  | uupdate v ((h as (v',old))::l) new =
		    if v = v'
		    then (v, case (new,old) of (Read        ,Write)        => WriteRead
			                     | (Read        ,WritePartial) => State       (* since R/W may concern different fields *)
				             | (Write       ,Read)         => State
				             | (WritePartial,Read)         => State
				             | (WriteRead   ,Read)         => State
					     | (Write       ,WritePartial) => Write
				             | (_           ,Unused)       => new
				             | _                           => old) :: l
		    else h :: uupdate v l new
		(* records variable usage in an expression *)
		fun expUsage all e m =
		    case e of
		        VarExp (v,ty) => uupdate v all m
  		      | InfixExp(_,_,e1,e2) => expUsage (expUsage all e1 m) e2 m
  		      | BinopExp(_,_,e1,e2) => expUsage (expUsage all e1 m) e2 m
  		      | UnopExp(_,_,e) => expUsage all e m
		      | RecordExp (fl,_) => listExpUsage all (List.map (fn (_,e) => e) fl) m
		      | ArrayExp (el,_) => listExpUsage all el m
		      | QueueExp (el,_) => listExpUsage all el m
		      | ArrayAccessExp (e1,e2,_) => expUsage (expUsage all e2 Read) e1 (case m of Write => WritePartial | _ => m)
		      | RecordAccessExp (e,_,_) => expUsage all e (case m of Write => WritePartial | _ => m)
		      | CondExp(c,e1,e2) => expUsage (expUsage (expUsage all c m) e1 m) e2 m
		      | LetExp(s,e) => expUsage (stUsage all s) e Read
		      | ConstrExp(_,e,_) => expUsage all e Read
		      | AppExp(_,el,_,_) => listExpUsage all el Read
		      | CallExp(_,el,_,_) => listExpUsage all el Read
		      | DestrExp (_,e) => expUsage all e Read
		      | MatchesExp (c,e) => expUsage all e Read
		      | _ => all
                and listExpUsage all [] m = all
		  | listExpUsage all (h::t) m = listExpUsage (expUsage all h m) t m
		(* records variable usage in a statement *)
		and stUsage all s =
		    case s of
			StSequence (s1,s2) => stUsage (stUsage all s1) s2
		      | StWhile(e,s) => let val alle = expUsage all e Read in join alle (stUsage alle s) end
		      | StForeach(v,s) => let val alle = expUsage all v Write in join alle (stUsage alle s) end
		      | StCond(e,s1,s2) => let val alle = expUsage all e Read in join (stUsage alle s1) (stUsage alle s2) end
		      | StAssign (p,e) => expUsage (expUsage all e Read) p Write   (* pattern matching assumed removed *)
		      | StGuard e => expUsage all e Read
		      | _ => all
		(* combines usage records for the branches of a conditional statement *)
		and join r1 r2 =
		    List.map (fn ((x,m1),(_,m2)) =>
		     	     let val m'=
		    	     	 if m1=m2 then m1 else
				 case (m1,m2) of
				     (State,_) => State
				   | (_,State) => State
				   | (Unused,Write) => WritePartial
				   | (Write,Unused) => WritePartial
				   | (Unused,m) => m
				   | (m,Unused) => m
				   | (Write,WritePartial) => WritePartial
				   | (WritePartial,Write) => WritePartial
				   | (Write,WriteRead) => WriteRead
				   | (WriteRead,Write) => WriteRead
				   | (WritePartial,WriteRead) => WritePartial
				   | (WriteRead,WritePartial) => WritePartial
				   (*
				   | (Read,Write) => State
				   | (Write,Read) => State
				   | (Read,WritePartial) => State
				   | (WritePartial,Read) => State
				   | (Read,WriteRead) => State
				   | (WriteRead,Read) => State
				    *)
				   | _ => State
				in (x,m')
				end) (ListPair.zip(r1,r2))
		(* records variable usage in a transition *)
	        fun trUsage (_,_,s,_,_) = merge (stUsage (map (fn (x,_) => (x,Unused)) (!accu)) s)
	    in (* analyze usage of variables *)
		List.app trUsage trl;
		let val (removed,left) = List.partition (fn (_,(_,Unused)) => true | (_,(_,Write)) => true | (_,(_,WritePartial)) => true
							  | _ => false) (ListPair.zip (vars,!accu))
		    val (newloc,left) = List.partition (fn (_,(_,State)) => true | _ => false) left
		    val (newtmp,left) = List.partition (fn (_,(_,Read)) => false | _ => true) left
		    val (inlined,newrtmp) = List.partition (fn ((_,ty,_),_) =>
							       case ty of BoolType => true
									| IntType => true | NatType => true | IntervalType _ => true
									| EnumType _ => true | UnionType _ => true
									| _ => false) left
		    val unused = map (fn ((s,_,_),_) => s) removed
                    (* for debug:
		    val _ =
			let open ErrPrint
			in print "----------------\n";
			   print "unused      ="; List.app (fn ((s,_,_),_) => (print s; print " ")) removed; print "\n"; flush ();
			   print "inlined     ="; List.app (fn ((s,_,_),_) => (print s; print " ")) inlined; print "\n"; flush ();
			   print "read only   ="; List.app (fn ((s,_,_),_) => (print s; print " ")) newrtmp; print "\n"; flush ();
			   print "temporaries ="; List.app (fn ((s,_,_),_) => (print s; print " ")) newtmp;  print "\n"; flush ();
			   print "state       ="; List.app (fn ((s,_,_),_) => (print s; print " ")) newloc;  print "\n"; flush ();
			   print "----------------\n"
			end
		    *)
		    (* inlines inlined variables and simplifies expressions [could be left to gcc but helps removing dead transitions] *)
		    exception DeadTrans
		    fun letExpTail (LetExp (_,e)) = letExpTail e | letExpTail e = e
 		    (* simplifies exp and removes assignements involving unused *)
 		    fun inlinee env e =
 			let val e' = expSimpl env e
 			in mapExp (fn e1 as LetExp (s,e2) =>
 				      (case mapStm (fn s as StAssign(p,e) =>
						      (* remove assignements of written-only variables *)
 						      if matchExp (fn VarExp (s,_) => mem s unused | _ => false) p then StNull else s
 						    | s => s) s of
 					  StNull => e2
 					| s' => LetExp(s',e2))
 				    | e => e) e'
 			end
 		    (* simplifies statement *)
		    fun inline env s =
			case s of
			    StGuard(e as LetExp _) => let val e' = inlinee env e
						      in case letExpTail e' of
							     BoolExp false => if !KEEPTRANS then StGuard e' else raise DeadTrans
							   | _ => StGuard e'
						      end
			  | StGuard e => (case inlinee env e of
						BoolExp true => StNull
					      | BoolExp false => if !KEEPTRANS then StGuard (BoolExp false) else raise DeadTrans
					      | e' => StGuard e')
			  | StSequence(s1,s2) => (case (inline env s1,inline env s2) of
						      (StNull,s2') => s2'
						    | (s1',StNull) => s1'
						    | s1s2' => StSequence s1s2')
			  | StWhile(e,s) => (case inlinee env e of
						 BoolExp true => comperror "loop detected in while statement"
					       | BoolExp false => inline env s
					       | e' => StWhile(e',inline env s))
			  | StForeach(v,s) => StForeach(v,inline env s)
			  | StCond(e,s1,s2) => (case inlinee env e of
						 BoolExp true => inline env s1
					       | BoolExp false => inline env s2
					       | e' => StCond(e',inline env s1,inline env s2))
			  | StAssign(p,e) => let val p' = inlinee env p
					     in (* remove assignements of written-only variables *)
						 if matchExp (fn VarExp (v,_) => mem v unused | _ => false) p'
						 then (REMOVED := true; StNull)
						 else StAssign (p',inlinee env e)
					     end
			  | StTo sl => StTo (isort sl String.>)
			  | _ => s
		    val inlinenv = map (fn ((s,_,v),_) => (s,v)) inlined
		in ProcessDecl (name,p,a,states,
				map fst newrtmp,
				map fst newtmp,
				map fst newloc,
				locports,prio,
				let val StTo e = init in StTo (isort e String.>) end,
				List.mapPartial (fn (tg,fr,s,b,prl) =>
					SOME (tg,isort fr String.>,inline inlinenv s,b,prl) handle
						    DeadTrans => (REMOVED := true; DEADTRANS := (hd (fst tg),ref prl) :: !DEADTRANS; NONE)) trl)
		end handle BYE => raise BYE | _ => failwith "moveVariables"
	    end
	  | moveVariables d = d

	(* apply recursively until no transition or statement removed *)
	fun recMove d = (REMOVED := false;
			 let val d' = moveVariables d
			 in if !REMOVED then recMove d' else d'
			 end);

    in recMove d
    end
  | simplify d = d;



(* removes equal (and removable) transitions [shares DEADTRANS with simplify] *)
(* ### currently does not check non-circularity for kept transitions (tina does it) ### *)
fun reduce (d as ProcessDecl (name,p,a,states,rtmpvars,tmpvars,locvars,locports,prio,init,trl)) =
    let 
	(* removes equal tansitions [should be a separate pass] *)
        fun remove_equal_transitions (d as ProcessDecl (name,p,a,states,rtmpvars,tmpvars,locvars,locports,prio,init,trl')) =
	    if !KEEPTRANS then d else 
	    let fun toInt b = Real.trunc (b * (pow (10.0,!DIGITS)))
		fun eqBounds (Infinite,Infinite) = true
		  | eqBounds (Closed b1,Closed b2) = toInt b1 = toInt b2
		  | eqBounds (Open b1,Open b2) = toInt b1 = toInt b2
		  | eqBounds _ = false
		fun islooptrans (_,src,_,(b1,b2),pr) =
		    not (eqBounds (b1,Closed 0.0) andalso eqBounds(b2,Infinite)) andalso
		    List.exists (fn p => String.sub(p,String.size p - 1) = #"?") src
		fun eqlist l1 l2 =
		    List.all (fn x => mem x l2) l1 andalso List.all (fn x => mem x l1) l2
		fun eqsup trl g1 g2 =
		    List.all (fn (_,_,_,_,prl) => mem g1 prl = mem g2 prl) trl andalso
		    List.all (fn (_,ref prl) => mem g1 prl = mem g2 prl) (!DEADTRANS)
		fun eqtr trl (t1 as (([g1],l1),fr,s,(b1,b2),prl1)) (([g2],l2),fr',s',(b1',b2'),prl2) =
		    (* same src and tgt: *)
		    fr=fr' andalso s=s'
		       (* same label (jp special :-)) *)
		       andalso l1=l2	   
		       (* same time intervals: *)
		       andalso eqBounds(b1,b1') andalso eqBounds(b2,b2')
		       (* t' < t1 <=> t' < t2 /\ forall t. t < t1 <=> t < t2; using closures would avoid false negatives: *)
		       andalso eqlist prl1 prl2 andalso eqsup trl g1 g2
		       (* identical timed looptrans should be kept: *)
		       andalso not (islooptrans t1)
		(* removes duplicated transitions, updates DEADTRANS for prio relation cleanup *)
		fun mkset [] = []
		  | mkset (l as (h::t)) =
		    let val (out,kept) = List.partition (eqtr l h) t
		    in List.app (fn (([nm],_),_,_,_,prl) => DEADTRANS := (nm,ref prl) :: !DEADTRANS) out;
		       h :: mkset kept
		    end
	    in (* let open ErrPrint in print "removed=\n"; List.app (fn (g,_) => (print g; print " ")) (!DEADTRANS); print "\n"; flush () end; *)
		ProcessDecl (name,p,a,states,rtmpvars,tmpvars,locvars,locports,prio,init,mkset trl')
	    end

        (* transitions that have been removed may have been captured in priority relation.
	   This function propagates their priority constraints and removes them from the
           priority lists of the kept transitions. *)
	fun clean_priority_relation (d as ProcessDecl (name,p,a,states,rtmpvars,tmpvars,locvars,locports,prio,init,trl)) =
	    let val dead = map fst (!DEADTRANS)
		(* computes the closure of l, dead trans excluded *)
		fun closure g l =
		    let val (out,kept) = List.partition (fn x => mem x dead) l
		    in if null out then l else
		       union kept (closure g (runion (map (fn x =>
		       	     if x=g then comperror "priority relation is cyclic" else !(assoc x (!DEADTRANS))) out)))
		    end
		(* replaces dead trans tags by their closure *)
		fun clean_prl (tr as (tg,fr,s,b,prl)) =
		    let val (dd,ok) = List.partition (fn g => mem g dead) prl
		    in if null dd then tr else (tg,fr,s,b,union ok (runion (map (fn g => !(assoc g (!DEADTRANS))) dd)))
		    end
	    in if null dead then d else
	       (List.app (fn (g,r as ref l) => r := closure g l) (!DEADTRANS);
		ProcessDecl (name,p,a,states,rtmpvars,tmpvars,locvars,locports,prio,init,map clean_prl trl))
	    end

	(* computes transitive closure of priority relation and checks nocycle *)
	fun check_prio_closure (d as ProcessDecl (_,_,_,_,_,_,_,_,_,_,trl)) =
	    if !NOPRIOCHECK then d else
	    let val (havingSucc,havingPred) =
		    ListPair.unzip (List.mapPartial (fn (_,_,_,_,[]) => NONE | (([tag],_),_,_,_,prl) => SOME (tag,prl)) trl)
	    in case List.filter (fn tag => List.exists (fn trset => mem tag trset) havingPred) havingSucc of
		   [] => d
		 | tags => (* tags of transitions having both a pred and succ in the priority relation *)
		   let val k = List.length tags
		       val table = Array.tabulate(k,fn _ => BitArray.array (k,false))
		       val updated = ref true
		       fun prod line j z accu = (* scalar product *)
			   if accu then (updated := true; BitArray.update(line,j,accu)) else
			   if z >= k then () else
			   prod line j (z+1) (accu orelse (BitArray.sub(line,z) andalso BitArray.sub(Array.sub(table,z),j)))
		       fun labof x ((([tag],lab),_,_,_,_)::trl) = if x=tag then lab else labof x trl
		   in (* initialize table so that table[i,j] <=> i < j *)
		       (* let open OutPrint in print "k = "; printi k; print "\n" end; *)
		       List.app (fn (_,_,_,_,[]) => ()
				  | (([tag],_),_,_,_,prl) =>
				    let val line = Array.sub(table,index tag tags)
				    in List.app (fn tr => BitArray.update(line,index tr tags,true)
						    handle Match => () (* tr not involved *)) prl
				    end
				    handle Match => () (* tag not involved *))
				trl;
		       (* compute closure *)
		       while !updated do (updated := false; 
		  			  iter k (fn i => let val line = Array.sub(table,i)
							  in iter k (fn j => if BitArray.sub(line,j) then () else prod line j 0 false)
							  end));
		       (* checks absence of cycles in priority relation *)
		       iter k (fn i => if BitArray.sub(Array.sub(table,i),i)
				       then let val lab = labof (List.nth(tags,i)) trl
					    in comperror ("priority relation found cyclic ("^lab^" < "^lab^")")
					    end
				       else ());
                       d
		   end
	    end

    in check_prio_closure (clean_priority_relation (remove_equal_transitions d))
    end
  | reduce d = d;


(* normalizes usage of explicit structures (ArrayExp/RecordExp/QueueExp and primitives returning queues) in
   statements and expressions for C generation. Introduces StBlock *)
local
    val UNIQUE = ref 0;
    (* new safe since all fiacre variables contains a "_" : *)
    fun new s = let val u = !UNIQUE in UNIQUE := 1 + !UNIQUE; s ^ Int.toString u end;
    fun zap () = UNIQUE := 0
    fun normSt s =
	let val decs = ref []
	    fun strip (LetExp(s,e)) = strip e | strip e = e
	    fun fls s =
		case s of (* first three cases handled as special case in emit *)
		    StAssign(p,RecordExp (fl,ty)) => StAssign(fle p,RecordExp (map (fn (f,e) => (f,fle e)) fl,ty))
		  | StAssign(p,ArrayExp(el,ty)) => StAssign(fle p,ArrayExp(map fle el,ty))
		  | StAssign(p,QueueExp(el,ty)) => StAssign(fle p,QueueExp(map fle el,ty))
		  | StAssign(p,e) => StAssign(fle p,fla e)
		  | StGuard e => StGuard(fle e)
		  | StSequence(s1,s2) => StSequence(normSt s1,normSt s2)
		  | StWhile(e,s) => StWhile(fle e,fls s)
		  | StForeach(v,s) => StForeach(fle v,fls s)
		  | StCond(c,s1,s2) => StCond(fle c,fls s1,fls s2)
		  | StReturn e => StReturn (fle e)
		  | _ => s
	    and fle e = (* e part of an expression. for structures, result must be a var;
                           also, arguments of queue primitives must be variables since queue macros duplicate them *)
		case e of
  		    BinopExp (ope,ty as QueueType(_,ety),e1,e2) => (* ope is enqueue or append *)
		    (case (e1,e2) of
			 (VarExp _, VarExp _) => let val q' = new "q"
						 in decs := (q',ty) :: !decs;
						    LetExp(StAssign(VarExp(q',ty),BinopExp (ope,ty,e1,e2)),
							   VarExp(q',ty))
						 end
		       | (VarExp _, _)        => let val e = new "e"
						     val q' = new "q"
						 in decs := (e,ety) :: (q',ty) :: !decs;
						    LetExp(StAssign(VarExp(e,ty),fla e2),
							   LetExp(StAssign(VarExp(q',ty),BinopExp (ope,ty,fle e1,VarExp(e,ety))),
								  VarExp(q',ty)))
						 end
		       | (_, VarExp _)        => let val q = new "q"
						     val q' = new "q"
						 in decs := (q,ty) :: (q',ty) :: !decs;
						    LetExp (StAssign(VarExp(q,ty),fla e1),
							    LetExp(StAssign(VarExp(q',ty),BinopExp (ope,ty,VarExp(q,ty),fle e2)),
								   VarExp(q',ty)))
						 end
		       | _                    => let val q = new "q"
						     val e = new "e"
						     val q' = new "q"
						 in decs := (q,ty) :: (e,ety) :: (q',ty) :: !decs;
						    LetExp (StAssign(VarExp(q,ty),fla e1),
							    LetExp(StAssign(VarExp(e,ty),fla e2),
								   LetExp(StAssign(VarExp(q',ty),BinopExp (ope,ty,VarExp(q,ty),VarExp(e,ety))),
									  VarExp(q',ty))))
						 end)
		  | UnopExp (DEQUEUE,ty,e1) =>
		    (case e1 of
			 VarExp _ => let val q' = new "q"
				     in decs := (q',ty) :: !decs;
					LetExp(StAssign(VarExp(q',ty),UnopExp (DEQUEUE,ty,fle e1)),
					       VarExp(q',ty))
				     end
		       | _        => let val q = new "q"
					 val q' = new "q"
				     in decs := (q,ty) :: (q',ty) :: !decs;
					LetExp(StAssign(VarExp(q,ty),fla e1),
					       LetExp(StAssign(VarExp(q',ty),UnopExp (DEQUEUE,ty,VarExp(q,ty))),
						      VarExp(q',ty)))
				     end)
		  | RecordExp (fl,ty) =>
		    let val r = new "r"
		    in decs := (r,ty) :: !decs;
		       LetExp (StAssign(VarExp(r,ty),RecordExp (map (fn (f,e) => (f,fle e)) fl,ty)),
			       VarExp (r,ty))
		    end
		  | ArrayExp (el,ty) =>
		    let val a = new "a"
		    in decs := (a,ty) :: !decs;
		       LetExp (StAssign(VarExp(a,ty),ArrayExp (map fle el,ty)),
			       VarExp (a,ty))
		    end
		  | QueueExp (el,ty as QueueType(sz,ety)) =>
		    let val q = new "q"
		    in decs := (q,ty) :: !decs;
		       LetExp (StAssign(VarExp(q,ty),QueueExp (map fle el,ty)),
			       VarExp (q,ty))
		    end
		  | ConExp (c,ty) =>
		    let val C = new "c"
		    in decs := (C,ty) :: !decs;
		       LetExp (StAssign(VarExp(C,ty),ConExp (c,ty)),
			       VarExp (C,ty))
		    end
		  | ConstrExp (c,e,ty) =>
		    let val C = new "c"
		    in decs := (C,ty) :: !decs;
		       LetExp (StAssign(VarExp(C,ty),ConstrExp(c,fle e,ty)),
			       VarExp (C,ty))
		    end
		  | UnopExp (ope,ty,e) => UnopExp (ope,ty,fle e)
		  | InfixExp (ope,ty,e1,e2) => InfixExp (ope,ty,fle e1,fle e2)
  		  | ArrayAccessExp (a,e,ty) => ArrayAccessExp (fle a,fle e,ty) 
		  | RecordAccessExp (r,f,ty) => RecordAccessExp (fle r,f,ty)
		  | CondExp (c,e1,e2) => CondExp (fle c,fle e1,fle e2) 
		  | LetExp (s,e) => LetExp (normSt s,fle e)
		  | DestrExp (c,e) => DestrExp (c,fle e)
		  | MatchesExp (c,e) => MatchesExp (c, fle e)
		  | AppExp (f,el,tyl,ty) => AppExp (f,map fle el,tyl,ty)
		  | CallExp (f,el,tyl,ty) => CallExp (f,map fle el,tyl,ty)
		  | _ => e
	    and fla e = (* e is rhs of an assignement *)
		case e of
  		    BinopExp (ope,ty as QueueType(_,ety),e1,e2) => (* ope is enqueue or append *)
		    (case (e1,e2) of
			 (VarExp _, VarExp _) => e
		       | (VarExp _, _)        => let val e = new "e"
						 in decs := (e,ety) :: !decs;
						    LetExp(StAssign(VarExp(e,ty),fla e2),
							   BinopExp (ope,ty,fle e1,VarExp(e,ety)))
						 end
		       | (_, VarExp _)        => let val q = new "q"
						 in decs := (q,ty) :: !decs;
						    LetExp (StAssign(VarExp(q,ty),fla e1),
							    BinopExp (ope,ty,VarExp(q,ty),fle e2))
						 end
		       | _                    => let val q = new "q"
						     val e = new "e"
						 in decs := (q,ty) :: (e,ety) :: !decs;
						    LetExp (StAssign(VarExp(q,ty),fla e1),
							    LetExp(StAssign(VarExp(e,ety),fla e2),
								   BinopExp (ope,ty,VarExp(q,ty),VarExp(e,ety))))
						 end)
		  | UnopExp (DEQUEUE,ty,e1) =>
		    (case e1 of
			 VarExp _ => e
		       | _        => let val q = new "q"
				     in decs := (q,ty) :: !decs;
					LetExp (StAssign(VarExp(q,ty),fla e1),
						UnopExp (DEQUEUE,ty,VarExp (q,ty)))
				     end)
		  | RecordExp (fl,ty) => RecordExp (map (fn (f,e) => (f,fle e)) fl,ty)
		  | ArrayExp (el,ty) => ArrayExp (map fle el,ty)
		  | QueueExp (el,ty as QueueType(sz,ety)) => QueueExp (map fle el,ty)
		  | ConstrExp (c,e,ty) => ConstrExp(c,fle e,ty)
		  | AppExp (f,el,tyl,ty) => AppExp (f,map fle el,tyl,ty)
		  | CallExp (f,el,tyl,ty) => CallExp (f,map fle el,tyl,ty)
		  | UnopExp (ope,ty,e) => UnopExp (ope,ty,fle e)
		  | InfixExp (ope,ty,e1,e2) => InfixExp (ope,ty,fle e1,fle e2)
  		  | ArrayAccessExp (a,e,ty) => ArrayAccessExp (fle a,fle e,ty) 
		  | RecordAccessExp (r,f,ty) => RecordAccessExp (fle r,f,ty)
		  | CondExp (c,e1,e2) => CondExp (fle c,fle e1,fle e2) 
		  | LetExp (s,e) => LetExp (normSt s,fle e)
		  | DestrExp (c,e) => DestrExp (c,fle e)
		  | MatchesExp (c,e) => MatchesExp (c, fle e)
		  | _ => e
	    val s' = fls s
	in if null (!decs) then s' else StBlock (rev (!decs),s')
	end;
in
fun normStructures (ProcessDecl(b,c,d,e,rk,k,f,h,i,init,g)) =
    (ProcessDecl(b,c,d,e,rk,k,f,h,i,init,map (fn (tag,fr,s,bnd,prl) => (zap (); (tag,fr,normSt s,bnd,prl))) g)
     handle BYE => raise BYE | _ => failwith "normStructures process")
  | normStructures (FunctionDecl(f,args,ty,vars,body)) = 
    let fun prefix s [] = s | prefix s ((x,ty,v)::t) = prefix (StSequence(StAssign(VarExp(x,ty),v),s)) t
    in FunctionDecl(f,args,ty,vars,normSt (prefix body (rev vars)))
    end
  | normStructures d = d;
end;


fun postcompose (Program dl) =
    Program (if !NOCOMP then dl else map (normStructures o reduce o simplify o normGuards o removePorts o insertTime) dl);

(*
fun postcompose (Program dl) =
    if !NOCOMP then Program dl else
    let val l1 = map insertTime dl handle _ => failwith "insertTime"
	val l2 = map removePorts l1 handle _ => failwith "removePorts"
	val l3 = map normGuards l2 handle _ => failwith "normGuards"
	val l4 = map simplify l3 handle _ => failwith "simplify"
	val l5 = map reduce l4 handle _ => failwith "reduce"
	val l6 = map normStructures l5 handle _ => failwith "normStructures"
    in Program l3
    end;
*)

end; (* PostCompose *)

